﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web.Mvc;
using RepPP.Toolkit;
using System.Web.Routing;

namespace RepPP.Toolkit.MVC2 {

    /// <summary>
    /// Class implementing a number of extension methods for Rep++ Filters
    /// </summary>
    public static class RepPPFilterExtensions {
        /// <summary>WildChar constants</summary>
        private const char      WILDCHAR_ANY       = '%';
        private const char      WILDCHAR_ONE       = '_';
        /// <summary>Suffix used to generate the name of the filter control</summary>
        private const string    BETWEEN_FLD_SUFFIX = "_Bet";       

        /// <summary>
        /// Gets the list of operators to be displayed in a filter
        /// </summary>
        /// <param name="htmlHelper">   Html Helper</param>
        /// <returns>
        /// List of operators
        /// </returns>
        public static SelectListItem[] GetFilterOperators(this HtmlHelper htmlHelper) {
            return new SelectListItem[] { new SelectListItem() { Text = "=",                                      Value = FilterOp.Equal.ToString(), Selected = true },
                                          new SelectListItem() { Text = "<",                                      Value = FilterOp.Less.ToString()},
                                          new SelectListItem() { Text = "<=",                                     Value = FilterOp.LessOrEqual.ToString()},
                                          new SelectListItem() { Text = ">",                                      Value = FilterOp.Greater.ToString()},
                                          new SelectListItem() { Text = ">=",                                     Value = FilterOp.GreaterOrEqual.ToString()},
                                          new SelectListItem() { Text = "<>",                                     Value = FilterOp.Different.ToString()},
                                          new SelectListItem() { Text = RepPPTranslator.RepPPText("between"),     Value = FilterOp.Between.ToString()},
                                          new SelectListItem() { Text = RepPPTranslator.RepPPText("is null"),     Value = FilterOp.IsNull.ToString()},
                                          new SelectListItem() { Text = RepPPTranslator.RepPPText("is not null"), Value = FilterOp.IsNotNull.ToString()}};
        }

        /// <summary>
        /// Generates the Html Markup for filter field
        /// </summary>
        /// <param name="htmlHelper">       Html Helper Object</param>
        /// <param name="fld">              Reference Field</param>
        /// <param name="fldOptions">       Field Control Options</param>
        /// <param name="htmlAttributes">   Html Attributes</param>
        /// <returns>
        /// The Field Control Html Markup
        /// </returns>
        public static MvcHtmlString RepPPFilterFld(this HtmlHelper      htmlHelper, 
                                                   RepPP.Field          fld, 
                                                   FieldControlOptions  fldOptions, 
                                                   object               htmlAttributes) {
            MvcHtmlString   strRetVal;
            FieldType       eType;

            eType = fld.FieldDef.Type;
            if (eType == FieldType.sdFieldSimpleAlphaChoice || eType == FieldType.sdFieldSimpleNumChoice) {
                strRetVal = htmlHelper.RepPPFldComboBox(fld, fldOptions, htmlAttributes);
            } else {
                strRetVal = htmlHelper.RepPPFldGeneric(fld, fldOptions, htmlAttributes);
            }
            return (strRetVal);
        }

        /// <summary>
        /// Gets the Field Html ID for the between control in a filter
        /// </summary>
        /// <param name="htmlHelper">   Html Helper Object</param>
        /// <param name="fld">          Rep++ Field Object</param>
        /// <returns>
        /// The Field Html ID 
        /// </returns>
        public static String RepPPGetFldBetweenHtmlID(this HtmlHelper htmlHelper, RepPP.Field fld) {
            return (Util.GetFldHtmlID(fld.FieldDef) + BETWEEN_FLD_SUFFIX);
        }

        /// <summary>
        /// Gets the Field Facade Html ID for the between control in a filter
        /// </summary>
        /// <param name="htmlHelper">   Html Helper Object</param>
        /// <param name="fld">          Rep++ Field Object</param>
        /// <returns>
        /// The Field Html ID 
        /// </returns>
        public static String RepPPGetFldBetweenFacadeHtmlID(this HtmlHelper htmlHelper, RepPP.Field fld) {
            return (Util.GetFldFacadeHtmlID(fld.FieldDef, new RouteValueDictionary(new { name = RepPPGetFldBetweenHtmlID(htmlHelper, fld) })));
        }

        /// <summary>
        /// Gets the operator values from the data that was posted back to the server
        /// </summary>
        /// <param name="rsdFilter">    Filter Rowset</param>
        /// <param name="formValues">   Posted back formvalues </param>
        /// <returns>
        /// Array representing operator values
        /// </returns>
        public static FilterOp[] GetFilterOperatorsValues(RowsetDef rsdFilter, FormCollection formValues) {
            FilterOp[]  arrRetVal;
            FieldDefs fldDefs;
            int       iCount;

            if (rsdFilter == null) {
                throw new ArgumentException("rsdFilter Cannot be null");
            }
            fldDefs   = rsdFilter.FieldDefs;
            iCount    = fldDefs.Count;
            arrRetVal = new FilterOp[iCount];
            for (int i = 0; i < iCount; i++ ) {
                try {
                    arrRetVal[i] = (FilterOp)Enum.Parse(typeof(FilterOp), formValues[fldDefs[i].Name + "_op"]);
                } catch (Exception) {
                    arrRetVal[i] = FilterOp.Equal;
                }
            }
            return (arrRetVal);
        }

        /// <summary>
        /// Gets the value of the And/Or operators
        /// </summary>
        /// <param name="formValues">   Posted back formvalues </param>
        /// <returns>
        /// An enumeration of the selected option
        /// </returns>
        public static FilterAndOr GetFilterAndOrValues(FormCollection formValues) {
            return ((string.Compare(formValues["AndOr"], "And", true) == 0 ? FilterAndOr.And : FilterAndOr.Or));
        }

        /// <summary>
        /// Gets the filter parameters using the filter RowsetDef and the
        /// posted form values
        /// </summary>
        /// <param name="rsdFilter">    Filter RowsetDef</param>
        /// <param name="formValues">   Posted form values</param>
        /// <param name="lstErrors">    List of validation errors</param>
        /// <returns>
        /// The Selection Buffer Filter Clause
        /// </returns>
        public static SBFilterClause GetFilterParameters(RowsetDef rsdFilter, FormCollection formValues, out Dictionary<string, string> lstErrors) {
            Application     app;
            SBFilterClause  sbFilterClauseRetVal = null;
            List<string>    lstParam1;
            List<string>    lstParam2;
            List<FilterOp>  lstOp;
            List<FieldDef>  lstFld;
            FilterAndOr     eAndOr;
            FilterOp        eOperator;
            FieldType       eFldType;
            bool            bParam1Empty;
            bool            bParam2Empty;

            app         = rsdFilter.Application;
            lstErrors   = new Dictionary<string, string>();
            lstFld      = rsdFilter.FieldDefs.ToList<FieldDef>();
            lstParam1   = Util.GetFieldsData(formValues, rsdFilter.FieldDefs, null /* strFldHtmlIDSuffix */);    
            ProcessValidationErrors(app.Errors, lstErrors, true /* Param 1*/);
            lstParam2   = Util.GetFieldsData(formValues, rsdFilter.FieldDefs, BETWEEN_FLD_SUFFIX);
            ProcessValidationErrors(app.Errors, lstErrors, false /* Param 2 */);
            lstOp       = GetFilterOperatorsValues(rsdFilter, formValues).ToList();
            eAndOr      = GetFilterAndOrValues(formValues);
            for (int i = rsdFilter.FieldDefs.Count -1; i >= 0; i--) {
                eOperator = lstOp[i];
                switch (eOperator) {
                case FilterOp.Between:
                    bParam1Empty = String.IsNullOrEmpty(lstParam1[i]);
                    bParam2Empty = String.IsNullOrEmpty(lstParam2[i]);
                    if (bParam1Empty || bParam2Empty) {
                        if (bParam1Empty) {
                            AddFldRequiredError(lstFld[i], lstErrors, true);
                        } 
                        if (bParam2Empty) {
                            AddFldRequiredError(lstFld[i], lstErrors, false);
                        }
                        lstParam1.RemoveAt(i);
                        lstParam2.RemoveAt(i);
                        lstOp.RemoveAt(i);
                        lstFld.RemoveAt(i);
                    } else {
                        lstParam1.Insert(i + 1, lstParam2[i]);
                    }
                    break;
                case FilterOp.IsNull:
                case FilterOp.IsNotNull:
                    lstParam1.RemoveAt(i);
                    lstParam2.RemoveAt(i);
                    break;
                default:
                    bParam1Empty = String.IsNullOrEmpty(lstParam1[i]);
                    if (bParam1Empty) {
                        if (eOperator != FilterOp.Equal) {
                            AddFldRequiredError(lstFld[i], lstErrors, true);
                        }
                        lstParam1.RemoveAt(i);
                        lstParam2.RemoveAt(i);
                        lstOp.RemoveAt(i);
                        lstFld.RemoveAt(i);
                    } else {
                        eFldType = lstFld[i].Type;
                        if (eOperator == FilterOp.Equal || eOperator == FilterOp.Different) {
                            if (eFldType == FieldType.sdFieldString         || 
                                eFldType == FieldType.sdFieldUpperString    || 
                                eFldType == FieldType.sdFieldFormatString) {
                                if (lstParam1[i].IndexOfAny(new char[] { WILDCHAR_ANY, WILDCHAR_ONE }) != -1) {
                                    lstOp[i] = (eOperator == FilterOp.Equal) ? FilterOp.Like : FilterOp.NotLike;
                                }
                            }
                        }
                    }
                    break;
                }
            }
            sbFilterClauseRetVal = new SBFilterClause(lstFld.ToArray(), lstOp.ToArray(), lstParam1.ToArray(), eAndOr);
            return (sbFilterClauseRetVal);
        }

        /// <summary>
        /// Builds a list of errors from the Rep++ Errors Collection
        /// </summary>
        /// <param name="errs">         The Rep++ Errors Collection</param>
        /// <param name="lstErrors">    List of errors</param>
        /// <param name="bParam1">      True to indicate that the errors are for the first filter parameter</param>
        private static void ProcessValidationErrors(Errors errs, Dictionary<string, string> lstErrors, bool bParam1) {
            FieldDef    fldDef;
            string      strObjectID;
            string      strPrompt;

            foreach (Error err in errs) {
                if (err.Number != (int)ErrorCode.sdAnsRequired) {
                    if (err.ObjType == ObjType.sdFieldDef) {
                        fldDef      = err.Obj() as FieldDef;
                        strObjectID = Util.GetFldHtmlID(fldDef) + (bParam1 ? String.Empty : BETWEEN_FLD_SUFFIX);
                        strPrompt   = fldDef.Prompt + ": ";
                    } else {
                        strObjectID = String.Empty;
                        strPrompt   = String.Empty;
                    }
                    lstErrors.Add(strObjectID,
                                  RepPPInputExtensions.GetErrorMessagePrefix() + strPrompt + (err.Msg ?? RepPPTranslator.RepPPText("InvalidValue")));
                }
            }
        }

        /// <summary>
        /// Add a required value field error
        /// </summary>
        /// <param name="fldDef">       The required fieldDef</param>
        /// <param name="lstErrors">    List of errors</param>
        /// <param name="bParam1">      True to indicate that the errors are for the first filter parameter</param>
        private static void AddFldRequiredError(FieldDef fldDef, Dictionary<string, string> lstErrors, bool bParam1) {
            string strMsg;
            string strFldHtmlID;

            strFldHtmlID = Util.GetFldHtmlID(fldDef) + (bParam1 ? String.Empty : BETWEEN_FLD_SUFFIX);
            if (!lstErrors.ContainsKey(strFldHtmlID)) {
                strMsg = RepPPInputExtensions.GetErrorMessagePrefix() + fldDef.Prompt + ": " + fldDef.Application.GetAtomText("$ANSRE", AtomType.sdAtomErr);
                lstErrors.Add(strFldHtmlID, strMsg);
            }
        }
    } // Class RepPPFilterExtensions
} // Namespace RepPP.Toolkit.MVC2