using System;
using System.Collections.Generic;
using System.IO;
using System.IO.IsolatedStorage;
using System.Reflection;
using System.Text;
using System.Xml;

namespace RepPP.Toolkit {

    /// <summary>
    /// Hold the UI Configuration into a XML document
    /// </summary>
    public class UIConfigHolderBase {

        /// <summary>Dictionary associating object type with an adaptor</summary>
        static protected Dictionary<Type,IUIConfigAdaptor>  m_dictTypes;
        /// <summary>XML document containing the configuration</summary>
        protected XmlDocument                               m_xmlDoc;
        /// <summary>Filename containing the configuration</summary>
        protected string                                    m_strFileName;
        /// <summary>true to use isolated storage, false to use standard file</summary>
        protected bool                                      m_bUseIsolatedStorage;

        /// <summary>
        /// Static class constructor
        /// </summary>
        static UIConfigHolderBase() {
            m_dictTypes = new Dictionary<Type,IUIConfigAdaptor>(64);
        }

        /// <summary>
        /// Class constructor
        /// </summary>
        protected UIConfigHolderBase() {
            m_xmlDoc                = null;
            m_strFileName           = null;
            m_bUseIsolatedStorage   = false;
        }

        /// <summary>
        /// Class constructor
        /// </summary>
        protected UIConfigHolderBase(XmlDocument xmlDoc) : this() {
            XmlDocument = xmlDoc;
        }

        /// <summary>
        /// Class constructor
        /// </summary>
        /// <param name="strFileName">          Name of the file</param>
        /// <param name="bUseIsolatedStorage">  true to use isolated storage, false to use normal file</param>
        protected UIConfigHolderBase(string strFileName, bool bUseIsolatedStorage) {
            XmlDocument                 xmlDoc;
            IsolatedStorageFile         isoStore;
            IsolatedStorageFileStream   fileStream;
            TextReader                  reader;

            m_strFileName           = strFileName;
            xmlDoc                  = new System.Xml.XmlDocument();
            m_bUseIsolatedStorage   = bUseIsolatedStorage;
            if (!String.IsNullOrEmpty(strFileName)) {
                if (m_bUseIsolatedStorage) {
                    isoStore    = IsolatedStorageFile.GetStore(IsolatedStorageScope.User | IsolatedStorageScope.Assembly | IsolatedStorageScope.Domain, null, null);
                    if (isoStore != null) {
                        using(isoStore) {
                            if (isoStore.GetFileNames(strFileName).Length > 0) {
                                fileStream  = new IsolatedStorageFileStream(strFileName, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite, isoStore);
                                using(reader = new StreamReader(fileStream)) {
                                    try {
                                        xmlDoc.Load(reader);
                                    } catch(System.Exception) {}
                                }
                            }
                        }
                    }
                } else {
                    try {
                        if (System.IO.File.Exists(strFileName)) {
                            xmlDoc.Load(strFileName);
                        }
                    } catch(System.Exception) {
                    }
                }
            }
            XmlDocument = xmlDoc;
        }

        /// <summary>
        /// Class constructor
        /// </summary>
        /// <param name="strFileName">      Name of the file</param>
        protected UIConfigHolderBase(string strFileName) : this(strFileName, false) {
        }

        /// <summary>
        /// Class constructor
        /// </summary>
        /// <param name="reader">       Text reader</param>
        protected UIConfigHolderBase(System.IO.TextReader reader) : this() {
            XmlDocument xmlDoc;
            
            xmlDoc  = new System.Xml.XmlDocument();
            if (reader != null) {
                try {
                    xmlDoc.Load(reader);
                } catch(System.Exception) {}
            }
            XmlDocument = xmlDoc;
        }

        /// <summary>
        /// Register an adaptor for the specified type
        /// </summary>
        /// <param name="type">         Type</param>
        /// <param name="adaptor">      Adaptor to associate with the type</param>
        public static void RegisterAdaptor(Type type, IUIConfigAdaptor adaptor) {
            m_dictTypes.Add(type, adaptor);
        }

        /// <summary>
        /// Validates a path component
        /// </summary>
        /// <param name="strPath">  Path component to validate</param>
        /// <returns>
        /// New path component
        /// </returns>
        private static string ValidatePath(string strPath) {
            foreach (char c in Path.GetInvalidFileNameChars()) { 
                strPath = strPath.Replace(c, '_'); 
            }
            strPath = strPath.Replace(' ', '_');
            return(strPath);
        }

        /// <summary>
        /// Build a configuration file name in the application data directory
        /// </summary>
        /// <param name="assemblyExe">  Main executable assembly </param>
        /// <param name="strBaseName">  Editor container editor name</param>
        /// <returns>
        /// Place where to put the configuration file
        /// </returns>
       public static string GetConfigFileName(Assembly assemblyExe, string strBaseName) {
            string          strRetVal;
            string          strCompanyName;
            string          strProduct;
            string          strVersion;
            System.Version  ver;
            object[]        attrs;

            attrs   = assemblyExe.GetCustomAttributes(typeof(AssemblyCompanyAttribute), false);
            if (attrs != null && attrs.Length > 0) {
                strCompanyName = ((AssemblyCompanyAttribute)attrs[0]).Company;
                if (strCompanyName != null) {
                    strCompanyName = strCompanyName.Trim();
                }
            } else {
                strCompanyName = null;
            }
            if (String.IsNullOrEmpty(strCompanyName)) {
                strCompanyName = "ACME";
            } else {
                strCompanyName = ValidatePath(strCompanyName);
            }
            attrs   = assemblyExe.GetCustomAttributes(typeof(AssemblyProductAttribute), false);
            if (attrs != null && attrs.Length > 0) {
                strProduct = ((AssemblyProductAttribute)attrs[0]).Product;
                if (strProduct != null) {
                    strProduct = strProduct.Trim();
                }
            } else {
                strProduct = null;
            }
            if (strProduct == null) {
                strProduct = assemblyExe.GetName().Name;
            }
            strProduct  = ValidatePath(strProduct);
            ver         = assemblyExe.GetName().Version;
            strVersion  = string.Format("{0}.{1}.{2}.{3}", ver.Major, ver.Minor, ver.Build, ver.Revision);
            if (strBaseName.IndexOf('.') == -1) {
                strBaseName += ".cfg";
            }
            strBaseName = ValidatePath(strBaseName);
            strRetVal   = System.Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData);
            strRetVal   = Path.Combine(strRetVal, strCompanyName);
            strRetVal   = Path.Combine(strRetVal, strProduct);
            strRetVal   = Path.Combine(strRetVal, strVersion);
            if (!Directory.Exists(strRetVal)) { 
                Directory.CreateDirectory(strRetVal); 
            }
            strRetVal   = Path.Combine(strRetVal, strBaseName);
            return(strRetVal);
        }

        /// <summary>
        /// Save the configuration to the file
        /// </summary>
        /// <returns>
        /// true if succeed, false if not
        /// </returns>
        public bool SaveToFile() {
            bool                        bRetVal = false;
            XmlDocument                 xmlDoc;
            IsolatedStorageFile         isoStore;
            IsolatedStorageFileStream   fileStream;
            TextWriter                  writer;
            
            if (!String.IsNullOrEmpty(m_strFileName)) {
                xmlDoc  = XmlDocument;
                if (m_bUseIsolatedStorage) {
                    isoStore    = IsolatedStorageFile.GetStore(IsolatedStorageScope.User | IsolatedStorageScope.Assembly | IsolatedStorageScope.Domain, null, null);
                    if (isoStore != null) {
                        using(isoStore) {
                            fileStream  = new IsolatedStorageFileStream(m_strFileName, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.ReadWrite, isoStore);
                            using(writer = new StreamWriter(fileStream)) {
                                try {
                                    xmlDoc.Save(writer);
                                    bRetVal = true;
                                } catch(System.Exception) {}
                            }
                        }
                    }
                } else {
                    try {
                        xmlDoc.Save(m_strFileName);
                        bRetVal = true;
                    } catch(System.Exception) {
                    }
                }
            }
            return(bRetVal);
        }

        /// <summary>
        /// Save the configuration to a writer
        /// </summary>
        /// <returns>
        /// true if succeed, false if not
        /// </returns>
        public bool SaveToWriter(System.IO.TextWriter writer) {
            bool        bRetVal = true;
            XmlDocument xmlDoc;
            
            try {
                xmlDoc  = XmlDocument;
                xmlDoc.Save(writer);
            } catch(System.Exception) {
                bRetVal  = false;
            }
            return(bRetVal);
        }

        /// <summary>
        /// XML document containing the configuration
        /// </summary>
        public XmlDocument XmlDocument {
            get {
                return(m_xmlDoc);
            }
            protected set {
                XmlNode node;

                if (value.ChildNodes.Count == 0) {
                    node = value.CreateNode(XmlNodeType.Element, "REP", "");
                    if (node != null) {
                        value.AppendChild(node);
                    }
                }
                m_xmlDoc = value;
            }
        }

        /// <summary>
        /// Gets a node attribute
        /// </summary>
        /// <param name="node">         Node in which to search the child. null for root</param>
        /// <param name="strName">      Attribute name to find</param>
        /// <returns>
        /// Attribute or null if not found
        /// </returns>
        public string GetNodeAttribute(XmlNode node, string strName) {
            string      strRetVal = null;
            
            strName = strName.ToUpper();
            foreach (XmlAttribute attr in node.Attributes) {
                if (attr.Name.ToUpper() == strName) {
                    strRetVal = attr.Value;
                    break;
                }
            }
            return(strRetVal);
        }

        /// <summary>
        /// Gets a node attribute
        /// </summary>
        /// <param name="node">         Node in which to search the child. null for root</param>
        /// <param name="strName">      Attribute name to find</param>
        /// <param name="bDefaultValue">Default value</param>
        /// <returns>
        /// Attribute or default value if not found
        /// </returns>
        public bool GetNodeAttribute(XmlNode node, string strName, bool bDefaultValue) {
            bool    bRetVal = bDefaultValue;
            string  str;
            
            str = GetNodeAttribute(node, strName);
            if (str != null && str.Length != 0) {
                if (!Boolean.TryParse(str, out bRetVal)) {
                    bRetVal = bDefaultValue;
                }
            }
            return(bRetVal);
        }

        /// <summary>
        /// Gets a node attribute
        /// </summary>
        /// <param name="node">         Node in which to search the child. null for root</param>
        /// <param name="strName">      Attribute name to find</param>
        /// <param name="iDefaultValue">Default value</param>
        /// <returns>
        /// Attribute or default value if not found
        /// </returns>
        public int GetNodeAttribute(XmlNode node, string strName, int iDefaultValue) {
            int     iRetVal = iDefaultValue;
            string  str;
            
            str = GetNodeAttribute(node, strName);
            if (str != null && str.Length != 0) {
                if (!Int32.TryParse(str, out iRetVal)) {
                    iRetVal = iDefaultValue;
                }
            }
            return(iRetVal);
        }

        /// <summary>
        /// Gets a node attribute
        /// </summary>
        /// <param name="node">         Node in which to search the child. null for root</param>
        /// <param name="strName">      Attribute name to find</param>
        /// <param name="dDefaultValue">Default value</param>
        /// <returns>
        /// Attribute or default value if not found
        /// </returns>
        public double GetNodeAttribute(XmlNode node, string strName, double dDefaultValue) {
            double  dRetVal = dDefaultValue;
            string  str;
            
            str = GetNodeAttribute(node, strName);
            if (str != null && str.Length != 0) {
                if (!Double.TryParse(str, out dRetVal)) {
                    dRetVal = dDefaultValue;
                }
            }
            return(dRetVal);
        }

        /// <summary>
        /// Sets a node attribute
        /// </summary>
        /// <param name="node">         Node in which to search the child. null for root</param>
        /// <param name="strName">      Attribute name to find</param>
        /// <param name="strVal">       Attribute value to set</param>
        public void SetNodeAttribute(XmlNode node, string strName, string strVal) {
            bool            bFound = false;
            XmlAttribute    attrNew;
            
            strName = strName.ToUpper();
            foreach (XmlAttribute attr in node.Attributes) {
                if (attr.Name.ToUpper() == strName) {
                    attr.Value = strVal;
                    bFound     = true;
                    break;
                }
            }
            if (!bFound) {
                attrNew       = m_xmlDoc.CreateAttribute(strName);
                attrNew.Value = strVal;
                node.Attributes.Append(attrNew);
            }
        }

        /// <summary>
        /// Sets a node attribute
        /// </summary>
        /// <param name="node">         Node in which to search the child. null for root</param>
        /// <param name="strName">      Attribute name to find</param>
        /// <param name="bVal">         Attribute value to set</param>
        public void SetNodeAttribute(XmlNode node, string strName, bool bVal) {
            SetNodeAttribute(node, strName, bVal ? "1" : "0");
        }

        /// <summary>
        /// Sets a node attribute
        /// </summary>
        /// <param name="node">         Node in which to search the child. null for root</param>
        /// <param name="strName">      Attribute name to find</param>
        /// <param name="iVal">         Attribute value to set</param>
        public void SetNodeAttribute(XmlNode node, string strName, int iVal) {
            SetNodeAttribute(node, strName, iVal.ToString());
        }

        /// <summary>
        /// Sets a node attribute
        /// </summary>
        /// <param name="node">         Node in which to search the child. null for root</param>
        /// <param name="strName">      Attribute name to find</param>
        /// <param name="dVal">         Attribute value to set</param>
        public void SetNodeAttribute(XmlNode node, string strName, double dVal) {
            SetNodeAttribute(node, strName, dVal.ToString());
        }


        /// <summary>
        /// Find a node
        /// </summary>
        /// <param name="node">                 Node in which to search the child. null for root</param>
        /// <param name="strName">              Node name to find</param>
        /// <param name="bCreateIfNotFound">    Creates the node if not found</param>
        /// <returns>
        /// Node or null if document is not loaded.
        /// </returns>
        public XmlNode FindNode(XmlNode node, string strName, bool bCreateIfNotFound) {
            XmlNode     nodeRetVal = null;
            XmlNodeList nodeList;
            
            if (m_xmlDoc != null) {
                strName  = strName.ToUpper();
                nodeList = (node == null) ? m_xmlDoc.DocumentElement.ChildNodes : node.ChildNodes;
                foreach (XmlNode nodeChild in nodeList) {
                    if (nodeChild.Name.ToUpper() == strName) {
                        nodeRetVal = nodeChild;
                        break;
                    }
                }
                if (nodeRetVal == null && bCreateIfNotFound) {
                    nodeRetVal = m_xmlDoc.CreateNode(XmlNodeType.Element, strName, "");
                    if (node == null) {
                        m_xmlDoc.DocumentElement.AppendChild(nodeRetVal);
                    } else {
                        node.AppendChild(nodeRetVal);
                    }
                }
            }
            return(nodeRetVal);
        }

        /// <summary>
        /// Find a node using an attribute value
        /// </summary>
        /// <param name="node">                 Node in which to search the child. null for root</param>
        /// <param name="strName">              Node name to find</param>
        /// <param name="strAttrName">          Attribute name to find</param>
        /// <param name="strAttrValue">         Attribute value</param>
        /// <param name="bCreateIfNotFound">    true to create if not found</param>
        /// <returns>
        /// Node or null if document is not loaded.
        /// </returns>
        public XmlNode FindNodeUsingAttribute(XmlNode   node, 
                                              string    strName, 
                                              string    strAttrName,
                                              string    strAttrValue, 
                                              bool      bCreateIfNotFound) {
            XmlNode     nodeRetVal = null;
            XmlNodeList nodeList;
            
            if (m_xmlDoc != null) {
                strName  = strName.ToUpper();
                nodeList = (node == null) ? m_xmlDoc.DocumentElement.ChildNodes : node.ChildNodes;
                foreach (XmlNode nodeChild in nodeList) {
                    if (nodeChild.Name.ToUpper() == strName &&
                        GetNodeAttribute(nodeChild, strAttrName) == strAttrValue) {
                        nodeRetVal = nodeChild;
                        break;
                    }
                }
                if (nodeRetVal == null && bCreateIfNotFound) {
                    nodeRetVal = m_xmlDoc.CreateNode(XmlNodeType.Element, strName, "");
                    SetNodeAttribute(nodeRetVal, strAttrName, strAttrValue);
                    if (node == null) {
                        m_xmlDoc.DocumentElement.AppendChild(nodeRetVal);
                    } else {
                        node.AppendChild(nodeRetVal);
                    }
                }
            }
            return(nodeRetVal);
        }

#if (VC10)
        /// <summary>
        /// Gets the parent node (creating the ancester as needed)
        /// </summary>
        /// <param name="parentNodes">          Ancester list</param>
        /// <param name="bCreateIfNotFound">    true to create the node if not found</param>
        /// <returns>
        /// Parent node
        /// </returns>
        protected XmlNode GetParentNode(List<Tuple<string, XmlNode>> parentNodes, bool bCreateIfNotFound) {
            XmlNode                 nodeRetVal = null;
            int                     iIndex;
            int                     iCount;
            Tuple<string, XmlNode>  tuple;

            iCount      = parentNodes.Count;
            nodeRetVal  = parentNodes[0].Item2;
            iIndex      = 1;
            while (iIndex < iCount && nodeRetVal != null) {
                tuple   = parentNodes[iIndex];
                if (tuple.Item2 == null) {
                    nodeRetVal = FindNode(nodeRetVal, tuple.Item1, bCreateIfNotFound);
                    if (nodeRetVal != null) {
                        tuple               = new Tuple<string,XmlNode>(tuple.Item1, nodeRetVal);
                        parentNodes[iIndex] = tuple;
                    }
                } else {
                    nodeRetVal  = tuple.Item2;
                }
                iIndex++;
            }
            return(nodeRetVal);
        }
#endif

    } //End class UIConfigHolderBase
} //End namespace
